syntax = "proto3";

package bar;
option go_package = "github.com/lyft/protoc-gen-star/testdata/generated/multipackage/bar";

import "multipackage/bar/baz/quux.proto";

message Buzz {
    uint64 id     = 1;
    baz.Quux quux = 2;
}
