syntax = "proto3";

package foo;
option go_package = "github.com/lyft/protoc-gen-star/testdata/generated/multipackage/foo";

import "multipackage/bar/buzz.proto";

message Fizz {
    uint64 id           = 1;
    bar.Buzz three_five = 2;
}
